"""URLs for django_fido application."""
from __future__ import unicode_literals

from django.conf.urls import url
from django.views.generic import TemplateView

from .views import U2fAuthenticationRequestView, U2fAuthenticationView, U2fRegistrationRequestView, U2fRegistrationView

app_name = 'django_fido'
urlpatterns = [
    url('^registration/$', U2fRegistrationView.as_view(), name='registration'),
    url('^registration/u2f_request/$', U2fRegistrationRequestView.as_view(), name='u2f_registration_request'),
    url('^registration/done/$', TemplateView.as_view(template_name='django_fido/registration_done.html'),
        name='registration_done'),
    url('^authentication/$', U2fAuthenticationView.as_view(), name='authentication'),
    url('^authentication/u2f_request/$', U2fAuthenticationRequestView.as_view(), name='u2f_authentication_request'),
]



"""Views for U2F registration and login."""
from __future__ import unicode_literals

import logging
from abc import ABCMeta, abstractmethod
from copy import deepcopy

from django.conf import settings
from django.contrib.auth import authenticate, get_user_model, login
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.views import LoginView
from django.http import JsonResponse
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.encoding import force_text
from django.utils.six import with_metaclass
from django.utils.translation import ugettext_lazy as _
from django.views.generic import FormView, View
from six.moves import http_client
from six.moves.urllib.parse import urlsplit, urlunsplit
from u2flib_server import u2f

from .constants import (AUTHENTICATION_REQUEST_SESSION_KEY, AUTHENTICATION_USER_SESSION_KEY,
                        REGISTRATION_REQUEST_SESSION_KEY, U2F_AUTHENTICATION_REQUEST, U2F_REGISTRATION_REQUEST)
from .forms import U2fResponseForm
from .models import U2fDevice

_LOGGER = logging.getLogger(__name__)


class BaseU2fRequestView(with_metaclass(ABCMeta, View)):
    """Base view for U2F request views.

    @cvar session_key: Session key where the U2F request is stored.
    @cvar u2f_request_factory: Function which accepts `app_id` and `registered_keys` and returns U2F request.
    """

    session_key = None
    u2f_request_factory = None

    def get_app_id(self):
        """Return appId - base URL to the web."""
        # Correct `appId` for the web application doesn't have a path.
        # See https://developers.yubico.com/U2F/App_ID.html
        chunks = urlsplit(self.request.build_absolute_uri('/'))
        return urlunsplit((chunks.scheme, chunks.netloc, '', '', ''))

    @abstractmethod
    def get_user(self):
        """Return user which is subject of the request."""
        pass

    def create_u2f_request(self):
        """Create and return U2F request.

        @raise ValueError: If request can't be created.
        """
        user = self.get_user()
        assert user.is_authenticated, "User must not be anonymous for U2F requests."
        registered_keys = [key.get_registered_key() for key in user.u2f_devices.all()]
        try:
            return self.u2f_request_factory(self.get_app_id(), registered_keys)
        except ValueError as error:
            raise ValueError("Can't create U2F request: {}".format(error))

    def get(self, request, *args, **kwargs):
        """Return JSON with U2F request."""
        try:
            u2f_request = self.create_u2f_request()
        except ValueError as error:
            return JsonResponse({'error': force_text(error)}, status=http_client.BAD_REQUEST)
        self.request.session[self.session_key] = u2f_request

        # Avoid bug in python-u2flib-server - https://github.com/Yubico/python-u2flib-server/issues/45
        # Remove `publicKey`s from the request.
        u2f_request_out = deepcopy(u2f_request)
        for key in u2f_request_out['registeredKeys']:
            key.pop('publicKey', None)

        return JsonResponse(u2f_request_out)


class U2fRegistrationRequestView(LoginRequiredMixin, BaseU2fRequestView):
    """Returns registration request and stores it in session."""

    session_key = REGISTRATION_REQUEST_SESSION_KEY
    u2f_request_factory = staticmethod(u2f.begin_registration)

    def get_user(self):
        """Return user which is subject of the request."""
        return self.request.user


class U2fRegistrationView(LoginRequiredMixin, FormView):
    """
    View to register U2F key.

    @cvar title: View title.
    @cvar u2f_request_url: URL at which an U2F request is provided.
    @cvar u2f_request_type: U2F request type
    """

    form_class = U2fResponseForm
    template_name = 'django_fido/u2f_form.html'
    success_url = reverse_lazy('django_fido:registration_done')

    title = _("Register Universal 2nd Factor (U2F) key")
    u2f_request_url = reverse_lazy('django_fido:u2f_registration_request')
    u2f_request_type = U2F_REGISTRATION_REQUEST

    def form_valid(self, form):
        """Complete the registration process."""
        u2f_request = self.request.session.pop(REGISTRATION_REQUEST_SESSION_KEY, None)
        if u2f_request is None:
            form.add_error(None, _('Registration request not found.'))
            return self.form_invalid(form)

        u2f_response = form.cleaned_data['u2f_response']

        try:
            device, attestation_cert = u2f.complete_registration(u2f_request, u2f_response)
        except (TypeError, ValueError, KeyError) as error:
            _LOGGER.info("U2F registration failed with error: %r", error)
            form.add_error(None, _('Registration failed.'))
            return self.form_invalid(form)

        U2fDevice.objects.create(
            user=self.request.user, version=device['version'], key_handle=device['keyHandle'],
            public_key=device['publicKey'], app_id=device['appId'], transports=device['transports'],
            raw_attestation=attestation_cert)
        return super(U2fRegistrationView, self).form_valid(form)


class U2fAuthenticationViewMixin(object):
    """
    Mixin for U2F authentication views.

    Ensure user to be authenticated exists.
    """

    def get_user(self):
        """
        Return user which is to be authenticated.

        Return None, if no user could be found.
        """
        user_pk = self.request.session.get(AUTHENTICATION_USER_SESSION_KEY)
        if user_pk is None:
            return None
        return get_user_model().objects.get(pk=user_pk)

    def dispatch(self, request, *args, **kwargs):
        """Redirect to login, if user couldn't be found."""
        user = self.get_user()
        if user is None or not user.is_authenticated:
            return redirect(settings.LOGIN_URL)
        return super(U2fAuthenticationViewMixin, self).dispatch(request, *args, **kwargs)


class U2fAuthenticationRequestView(U2fAuthenticationViewMixin, BaseU2fRequestView):
    """Returns authentication request and stores it in session."""

    session_key = AUTHENTICATION_REQUEST_SESSION_KEY
    u2f_request_factory = staticmethod(u2f.begin_authentication)


class U2fAuthenticationView(U2fAuthenticationViewMixin, LoginView):
    """
    View to authenticate U2F key.

    @cvar title: View title.
    @cvar u2f_request_url: URL at which an U2F request is provided.
    @cvar u2f_request_type: U2F request type
    """

    form_class = U2fResponseForm
    template_name = 'django_fido/u2f_form.html'

    title = _("Authenticate Universal 2nd Factor (U2F) key")
    u2f_request_url = reverse_lazy('django_fido:u2f_authentication_request')
    u2f_request_type = U2F_AUTHENTICATION_REQUEST

    def get_form_kwargs(self):
        """Return form arguments - exclude request."""
        kwargs = super(U2fAuthenticationView, self).get_form_kwargs()
        # U2fResponseForm doesn't accept request.
        kwargs.pop('request', None)
        return kwargs

    def form_valid(self, form):
        """Complete the registration process."""
        u2f_request = self.request.session.pop(AUTHENTICATION_REQUEST_SESSION_KEY, None)
        if u2f_request is None:
            form.add_error(None, _('Authentication request not found.'))
            return self.form_invalid(form)

        u2f_response = form.cleaned_data['u2f_response']

        user = authenticate(request=self.request, user=self.get_user(), u2f_request=u2f_request,
                            u2f_response=u2f_response)
        if user is None:
            form.add_error(None, _('Authentication failed.'))
            return self.form_invalid(form)

        login(self.request, user)
        # Ensure user is deleted from session.
        self.request.session.pop(AUTHENTICATION_USER_SESSION_KEY, None)

        return redirect(self.get_success_url())



"""Test `django_fido.models`."""
from __future__ import unicode_literals

from django.core.exceptions import ValidationError
from django.test import SimpleTestCase
from mock import sentinel

from django_fido.models import TransportsValidator, U2fDevice


class TestTransportsValidator(SimpleTestCase):
    """Test `TransportsValidator`."""

    valid_values = (
        # Single values
        'bt', 'ble', 'nfc', 'usb', 'usb-internal',
        # Mutliple values
        'bt,ble', 'nfc,usb', 'bt,ble,nfc,usb,usb-internal',
    )

    def test_valid(self):
        validator = TransportsValidator()
        for value in self.valid_values:
            validator(value)

    # (input - invalid chunk) pairs
    invalid_values = (
        ('', ''),
        (' ', ' '),
        (',', ''),
        # Unknown value
        ('junk', 'junk'),
        # No extra spaces
        ('bt, ble', ' ble'),
        ('nfc ,usb', 'nfc '),
    )

    def test_invalid(self):
        validator = TransportsValidator()

        for value, invalid_chunk in self.invalid_values:
            with self.assertRaisesMessage(ValidationError, 'Select a valid choice.') as catcher:
                validator(value)
            self.assertEqual(catcher.exception.code, 'invalid_choice')
            self.assertEqual(catcher.exception.params, {'value': invalid_chunk})

    def test_custom_choices(self):
        validator = TransportsValidator(choices=['foo', 'bar'])

        validator('foo,bar')
        self.assertRaisesMessage(ValidationError, 'Select a valid choice.', validator, 'usb')

    def test_custom_code(self):
        validator = TransportsValidator(code='smeghead')

        with self.assertRaisesMessage(ValidationError, 'Select a valid choice.') as catcher:
            validator('rimmer')
        self.assertEqual(catcher.exception.code, 'smeghead')

    def test_custom_message(self):
        validator = TransportsValidator(message="You're a smeghead.")

        self.assertRaisesMessage(ValidationError, "You're a smeghead.", validator, 'rimmer')


class TestU2fDevice(SimpleTestCase):
    """Test `U2fDevice` model."""

    # serialized, deserialized pairs
    transport_values = (
        (None, []),
        ('usb', ['usb']),
        ('bt,ble', ['bt', 'ble']),
    )

    def test_transports_getter(self):
        for raw, full in self.transport_values:
            u2f_key = U2fDevice(raw_transports=raw)
            self.assertEqual(u2f_key.transports, full)

    def test_transports_setter(self):
        for raw, full in self.transport_values:
            u2f_key = U2fDevice()
            u2f_key.transports = full
            self.assertEqual(u2f_key.raw_transports, raw)

    # base64 encoded, decoded pairs
    attestation_values = (
        (None, None),
        ('UmltbWVyJ3Mgc2lsdmVyIHN3aW1taW5nIGNlcnRpZmljYXRl', b"Rimmer's silver swimming certificate"),
    )

    def test_raw_attestation_getter(self):
        for encoded, decoded in self.attestation_values:
            u2f_key = U2fDevice(attestation=encoded)
            self.assertEqual(u2f_key.raw_attestation, decoded)

    def test_raw_attestation_setter(self):
        for encoded, decoded in self.attestation_values:
            u2f_key = U2fDevice()
            u2f_key.raw_attestation = decoded
            self.assertEqual(u2f_key.attestation, encoded)

    def test_get_registered_key(self):
        u2f_key = U2fDevice(key_handle=sentinel.key_handle, app_id=sentinel.app_id, version=sentinel.version,
                            raw_transports='usb,nfs', public_key=sentinel.public_key)
        retgistered_key = {'keyHandle': sentinel.key_handle,
                           'publicKey': sentinel.public_key,
                           'appId': sentinel.app_id,
                           'version': sentinel.version,
                           'transports': ['usb', 'nfs']}
        self.assertEqual(u2f_key.get_registered_key(), retgistered_key)



"""Models for storing keys."""
from __future__ import unicode_literals

import base64

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.deconstruct import deconstructible
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

# https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-javascript-api-v1.2-ps-20170411.html#u2f-transports
TRANSPORT_CHOICES = (
    ('bt', _('Bluetooth Classic (Bluetooth BR/EDR)')),
    ('ble', _('Bluetooth Low Energy (Bluetooth Smart)')),
    ('nfc', _('Near-Field Communications')),
    ('usb', _('USB HID')),
    ('usb-internal', _('Non-removable USB HID')),
)


@deconstructible
class TransportsValidator(object):
    """Validator for comma separated transport values.

    @ivar choices: List/tuple of available values.
    """

    choices = tuple(choice for choice, label in TRANSPORT_CHOICES)
    code = 'invalid_choice'
    message = _('Select a valid choice. %(value)s is not one of the available choices.')

    def __init__(self, choices=None, code=None, message=None):
        """Set custom `choices`, `code` or `message`."""
        if choices is not None:
            self.choices = choices
        if code is not None:
            self.code = code
        if message is not None:
            self.message = message

    def __call__(self, value):
        """Validate the input."""
        for chunk in force_text(value).split(','):
            if chunk not in self.choices:
                raise ValidationError(self.message, code=self.code, params={'value': chunk})


class U2fDevice(models.Model):
    """Represents a registered U2F device.

    U2F registered key fields:
     * version
     * key_handle
     * public_key
     * app_id
     * raw_transports - comma separated list of transports
     * attestation - Base64 encoded attestation certificate (X509 in DER format)

    Authentication fields:
     * counter
    """

    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='u2f_devices', on_delete=models.CASCADE)
    create_datetime = models.DateTimeField(auto_now_add=True)

    version = models.TextField()
    key_handle = models.TextField()
    public_key = models.TextField()
    app_id = models.TextField(blank=True, null=True, default=None)
    raw_transports = models.TextField(blank=True, null=True, default=None, validators=[TransportsValidator()])
    attestation = models.TextField(blank=True, null=True, default=None)

    counter = models.PositiveIntegerField(default=0)

    @property
    def transports(self):
        """Return U2F transports."""
        if not self.raw_transports:
            return []
        return self.raw_transports.split(',')

    @transports.setter
    def transports(self, value):
        if not value:
            self.raw_transports = None
        else:
            self.raw_transports = ','.join(value)

    @property
    def raw_attestation(self):
        """Return attestation certificate in DER format.

        @rtype: bytes
        """
        if self.attestation is None:
            return None
        return base64.b64decode(self.attestation)

    @raw_attestation.setter
    def raw_attestation(self, value):
        if value is None:
            self.attestation = None
        else:
            # Encode to base64 and the decode to six.text_type
            self.attestation = base64.b64encode(value).decode('utf-8')

    def get_registered_key(self):
        """Return data for `RegisteredKey` structure.

        See https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-javascript-api-v1.2-ps-20170411.html
        """
        return {
            'keyHandle': self.key_handle,
            'appId': self.app_id,
            'version': self.version,
            'transports': self.transports,
            # Avoid bug in python-u2flib-server - https://github.com/Yubico/python-u2flib-server/issues/45
            'publicKey': self.public_key,
        }



"""
Django FIDO constants.

Session keys to store U2F requests:
 * REGISTRATION_REQUEST_SESSION_KEY
 * AUTHENTICATION_REQUEST_SESSION_KEY

Session key to store user PK for U2F authentication
 * AUTHENTICATION_USER_SESSION_KEY

U2F request type identifiers:
 * U2F_REGISTRATION_REQUEST
 * U2F_AUTHENTICATION_REQUEST
These are shared between code and JS.
"""
from __future__ import unicode_literals

REGISTRATION_REQUEST_SESSION_KEY = 'fido_u2f_registration_request'
AUTHENTICATION_REQUEST_SESSION_KEY = 'fido_u2f_authentication_request'

AUTHENTICATION_USER_SESSION_KEY = 'fido_u2f_user'

U2F_REGISTRATION_REQUEST = 'registration'
U2F_AUTHENTICATION_REQUEST = 'authentication'



"""Forms for U2F registration and login."""
from __future__ import unicode_literals

import json

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _


class U2fResponseForm(forms.Form):
    """Form for U2F responses."""

    u2f_response = forms.CharField(error_messages={'required': _("Operation wasn't completed.")},
                                   widget=forms.HiddenInput)

    class Media:
        """Add U2F related JS."""

        js = ('django_fido/js/u2f-api.js', 'django_fido/js/u2f-registration.js')

    def clean_u2f_response(self):
        """Ensure U2F response is valid JSON."""
        u2f_response = self.cleaned_data['u2f_response']
        try:
            return json.loads(u2f_response)
        except ValueError:
            raise ValidationError(_('U2F response is malformed.'), code='invalid')



"""Test `django_fido.views` module."""
from __future__ import unicode_literals

import json

from django.contrib.auth import get_user, get_user_model
from django.core.exceptions import NON_FIELD_ERRORS
from django.test import TestCase, override_settings
from django.urls import reverse, reverse_lazy

from django_fido.models import U2fDevice
from django_fido.views import (AUTHENTICATION_REQUEST_SESSION_KEY, AUTHENTICATION_USER_SESSION_KEY,
                               REGISTRATION_REQUEST_SESSION_KEY)

from .utils import TEMPLATES

User = get_user_model()


@override_settings(ROOT_URLCONF='django_fido.tests.urls')
class TestU2fRegistrationRequestView(TestCase):
    """Test `U2fRegistrationRequestView` class."""

    url = reverse_lazy('django_fido:u2f_registration_request')
    session_key = REGISTRATION_REQUEST_SESSION_KEY

    def test_anonymous(self):
        with self.settings(LOGIN_URL='/login/'):
            response = self.client.get(self.url)

        self.assertRedirects(response, '/login/?next={}'.format(self.url), fetch_redirect_response=False)

    def test_get(self):
        user = User.objects.create_user('kryten')
        self.client.force_login(user)

        response = self.client.get(self.url)

        self.assertEqual(response.status_code, 200)
        # Check response contains the same request as session
        u2f_request = self.client.session[self.session_key]
        self.assertEqual(response.json(), u2f_request)
        # Check request structure
        self.assertEqual(u2f_request['appId'], 'http://testserver')
        self.assertEqual(u2f_request['registeredKeys'], [])
        self.assertEqual(u2f_request['registerRequests'][0]['version'], 'U2F_V2')
        self.assertIn('challenge', u2f_request['registerRequests'][0])

    def test_get_registered_keys(self):
        user = User.objects.create_user('kryten')
        U2fDevice.objects.create(user=user, version='42', key_handle='Left nipple', public_key='Yes', app_id='FM Radio',
                                 transports=['ble'])
        self.client.force_login(user)

        response = self.client.get(self.url)

        self.assertEqual(response.status_code, 200)
        # Check request structure
        registered_key = {'keyHandle': 'Left nipple', 'publicKey': 'Yes', 'appId': 'FM Radio', 'version': '42',
                          'transports': ['ble']}
        u2f_request = self.client.session[self.session_key]
        self.assertEqual(u2f_request['appId'], 'http://testserver')
        self.assertEqual(u2f_request['registeredKeys'], [registered_key])
        self.assertEqual(u2f_request['registerRequests'][0]['version'], 'U2F_V2')
        self.assertIn('challenge', u2f_request['registerRequests'][0])
        # Check response contains the same request as session (except for the public key).
        public_registered_key = {'keyHandle': 'Left nipple', 'appId': 'FM Radio', 'version': '42',
                                 'transports': ['ble']}
        response_data = {'appId': 'http://testserver', 'registerRequests': u2f_request['registerRequests'],
                         'registeredKeys': [public_registered_key]}
        self.assertEqual(response.json(), response_data)


@override_settings(ROOT_URLCONF='django_fido.tests.urls', TEMPLATES=TEMPLATES)
class TestU2fRegistrationView(TestCase):
    """Test `U2fRegistrationView` class."""

    url = reverse_lazy('django_fido:registration')
    u2f_request = {'appId': 'http://testserver', 'registeredKeys': [],
                   'registerRequests': [{'challenge': 'Listers_underpants', 'version': 'U2F_V2'}]}
    u2f_response = {
        'registrationData': ('BQR4cOSm5qgX1VDRogGXslsVW6wbVI04ccE3g3D5nBvITulx8ad26J6fzhFR5A-tNYUzCB-nEXmC1dagSMidDXiFQ'
                             'ImCDvzeNu-hzQ_1-SFfEj-xYvW0HMXSHjh7tq_cH-0bhbqfcsAfl8FNfGJGNAlrB_O9Pqm2XcO6uSbs8hOBqkEwgg'
                             'JEMIIBLqADAgECAgR4wN8OMAsGCSqGSIb3DQEBCzAuMSwwKgYDVQQDEyNZdWJpY28gVTJGIFJvb3QgQ0EgU2VyaWF'
                             'sIDQ1NzIwMDYzMTAgFw0xNDA4MDEwMDAwMDBaGA8yMDUwMDkwNDAwMDAwMFowKjEoMCYGA1UEAwwfWXViaWNvIFUy'
                             'RiBFRSBTZXJpYWwgMjAyNTkwNTkzNDBZMBMGByqGSM49AgEGCCqGSM49AwEHA0IABLW4cVyD_f4OoVxFd6yFjfSMF'
                             '2_eh53K9Lg9QNMg8m-t5iX89_XIr9g1GPjbniHsCDsYRYDHF-xKRwuWim-6P2-jOzA5MCIGCSsGAQQBgsQKAgQVMS'
                             '4zLjYuMS40LjEuNDE0ODIuMS4xMBMGCysGAQQBguUcAgEBBAQDAgUgMAsGCSqGSIb3DQEBCwOCAQEAPvar9kqRawv'
                             '5lJON3JU04FRAAmhWeKcsQ6er5l2QZf9h9FHOijru2GaJ0ZC5UK8AelTRMe7wb-JrTqe7PjK3kgWl36dgBDRT40r4'
                             'RMN81KhfjFwthw4KKLK37UQCQf2zeSsgdrDhivqbQy7u_CZYugkFxBskqTxuyLum1W8z6NZT189r1QFUVaJll0D33'
                             'MUcwDFgnNA-ps3pOZ7KCHYykHY_tMjQD1aQaaElSQBq67BqIaIU5JmYN7Qp6B1-VtM6VJLdOhYcgpOVQIGqfu90nD'
                             'pWPb3X26OVzEc-RGltQZGFwkN6yDrAZMHL5HIn_3obd8fV6gw2fUX2ML2ZjVmybjBEAiBTOUwY12wm2TrRMsAs-EK'
                             'PacouX_X7bOEz6vnLk03JtAIgaqR9H4lOx4JBeQb-xwLPz1shKfxwD1pVY67X-m8ACEI'),
        'clientData': ('eyAiY2hhbGxlbmdlIjogIkxpc3RlcnNfdW5kZXJwYW50cyIsICJvcmlnaW4iOiAiaHR0cDpcL1wvbW9qZWlkLnZ6aW1hIiw'
                       'gInR5cCI6ICJuYXZpZ2F0b3IuaWQuZmluaXNoRW5yb2xsbWVudCIgfQ'),
        'version': 'U2F_V2'}

    attestation = (
        'MIICRDCCAS6gAwIBAgIEeMDfDjALBgkqhkiG9w0BAQswLjEsMCoGA1UEAxMjWXViaWNvIFUyRiBSb290IENBIFNlcmlhbCA0NTcyMDA2MzEwIB'
        'cNMTQwODAxMDAwMDAwWhgPMjA1MDA5MDQwMDAwMDBaMCoxKDAmBgNVBAMMH1l1YmljbyBVMkYgRUUgU2VyaWFsIDIwMjU5MDU5MzQwWTATBgcq'
        'hkjOPQIBBggqhkjOPQMBBwNCAAS1uHFcg/3+DqFcRXeshY30jBdv3oedyvS4PUDTIPJvreYl/Pf1yK/YNRj4254h7Ag7GEWAxxfsSkcLlopvuj'
        '9vozswOTAiBgkrBgEEAYLECgIEFTEuMy42LjEuNC4xLjQxNDgyLjEuMTATBgsrBgEEAYLlHAIBAQQEAwIFIDALBgkqhkiG9w0BAQsDggEBAD72'
        'q/ZKkWsL+ZSTjdyVNOBUQAJoVninLEOnq+ZdkGX/YfRRzoo67thmidGQuVCvAHpU0THu8G/ia06nuz4yt5IFpd+nYAQ0U+NK+ETDfNSoX4xcLY'
        'cOCiiyt+1EAkH9s3krIHaw4Yr6m0Mu7vwmWLoJBcQbJKk8bsi7ptVvM+jWU9fPa9UBVFWiZZdA99zFHMAxYJzQPqbN6Tmeygh2MpB2P7TI0A9W'
        'kGmhJUkAauuwaiGiFOSZmDe0KegdflbTOlSS3ToWHIKTlUCBqn7vdJw6Vj2919ujlcxHPkRpbUGRhcJDesg6wGTBy+RyJ/96G3fH1eoMNn1F9j'
        'C9mY1Zsm4=')

    def test_anonymous(self):
        with self.settings(LOGIN_URL='/login/'):
            response = self.client.get(self.url)

        self.assertRedirects(response, '/login/?next={}'.format(self.url), fetch_redirect_response=False)

    def test_get(self):
        user = User.objects.create_user('kryten')
        self.client.force_login(user)

        response = self.client.get(self.url)

        self.assertContains(response, 'Register Universal 2nd Factor (U2F) key')

    def test_post(self):
        user = User.objects.create_user('kryten')
        self.client.force_login(user)
        session = self.client.session
        session[REGISTRATION_REQUEST_SESSION_KEY] = self.u2f_request
        session.save()

        response = self.client.post(self.url, {'u2f_response': json.dumps(self.u2f_response)})

        self.assertRedirects(response, reverse('django_fido:registration_done'))
        queryset = U2fDevice.objects.values_list('user__pk', 'version', 'key_handle', 'public_key', 'app_id',
                                                 'raw_transports', 'attestation')
        key_data = (user.pk, 'U2F_V2',
                    'iYIO_N4276HND_X5IV8SP7Fi9bQcxdIeOHu2r9wf7RuFup9ywB-XwU18YkY0CWsH870-qbZdw7q5JuzyE4GqQQ',
                    'BHhw5KbmqBfVUNGiAZeyWxVbrBtUjThxwTeDcPmcG8hO6XHxp3bonp_OEVHkD601hTMIH6cReYLV1qBIyJ0NeIU',
                    'http://testserver', 'usb', self.attestation)
        self.assertQuerysetEqual(queryset, [key_data], transform=tuple)
        self.assertNotIn(REGISTRATION_REQUEST_SESSION_KEY, self.client.session)

    def test_post_no_session(self):
        user = User.objects.create_user('kryten')
        self.client.force_login(user)

        response = self.client.post(self.url, {'u2f_response': 'null'})

        self.assertContains(response, 'Register Universal 2nd Factor (U2F) key')
        self.assertEqual(response.context['form'].errors, {NON_FIELD_ERRORS: ['Registration request not found.']})
        self.assertNotIn(REGISTRATION_REQUEST_SESSION_KEY, self.client.session)

    def test_post_invalid_response(self):
        user = User.objects.create_user('kryten')
        self.client.force_login(user)
        session = self.client.session
        session[REGISTRATION_REQUEST_SESSION_KEY] = self.u2f_request
        session.save()

        response = self.client.post(self.url, {'u2f_response': 'null'})

        self.assertContains(response, 'Register Universal 2nd Factor (U2F) key')
        self.assertEqual(response.context['form'].errors, {NON_FIELD_ERRORS: ['Registration failed.']})
        self.assertNotIn(REGISTRATION_REQUEST_SESSION_KEY, self.client.session)


@override_settings(ROOT_URLCONF='django_fido.tests.urls')
class TestU2fAuthenticationRequestView(TestCase):
    """Test `U2fAuthenticationRequestView` class."""

    url = reverse_lazy('django_fido:u2f_authentication_request')
    session_key = AUTHENTICATION_REQUEST_SESSION_KEY

    def test_no_user(self):
        with self.settings(LOGIN_URL='/login/'):
            response = self.client.get(self.url)

        self.assertRedirects(response, '/login/', fetch_redirect_response=False)

    def test_get(self):
        user = User.objects.create_user('kryten')
        U2fDevice.objects.create(user=user, version='42', key_handle='Left nipple', public_key='Yes', app_id='FM Radio',
                                 transports=['ble'])
        session = self.client.session
        session[AUTHENTICATION_USER_SESSION_KEY] = user.pk
        session.save()

        response = self.client.get(self.url)

        self.assertEqual(response.status_code, 200)
        # Check request structure
        u2f_request = self.client.session[self.session_key]
        self.assertEqual(u2f_request['appId'], 'http://testserver')
        registered_key = {'keyHandle': 'Left nipple', 'publicKey': 'Yes', 'appId': 'FM Radio', 'version': '42',
                          'transports': ['ble']}
        self.assertEqual(u2f_request['registeredKeys'], [registered_key])
        self.assertIn('challenge', u2f_request)
        # Check response contains the same request as session (except for the public key).
        public_registered_key = {'keyHandle': 'Left nipple', 'appId': 'FM Radio', 'version': '42',
                                 'transports': ['ble']}
        response_data = {'appId': 'http://testserver', 'registeredKeys': [public_registered_key],
                         'challenge': u2f_request['challenge']}
        self.assertEqual(response.json(), response_data)

    def test_get_no_keys(self):
        user = User.objects.create_user('kryten')
        session = self.client.session
        session[AUTHENTICATION_USER_SESSION_KEY] = user.pk
        session.save()

        response = self.client.get(self.url)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {'error': "Can't create U2F request: Must have at least one RegisteredKey"})
        self.assertNotIn(self.session_key, self.client.session)


@override_settings(ROOT_URLCONF='django_fido.tests.urls', TEMPLATES=TEMPLATES,
                   AUTHENTICATION_BACKENDS=['django_fido.backends.U2fAuthenticationBackend'])
class TestU2fAuthenticationView(TestCase):
    """Test `U2fAuthenticationView` class."""

    url = reverse_lazy('django_fido:authentication')
    # Valid U2F data
    key_handle = 'iYIO_N4276HND_X5IV8SP7Fi9bQcxdIeOHu2r9wf7RuFup9ywB-XwU18YkY0CWsH870-qbZdw7q5JuzyE4GqQQ'
    public_key = 'BHhw5KbmqBfVUNGiAZeyWxVbrBtUjThxwTeDcPmcG8hO6XHxp3bonp_OEVHkD601hTMIH6cReYLV1qBIyJ0NeIU'
    registered_key = {'publicKey': public_key, 'keyHandle': key_handle, 'version': 'U2F_V2', 'transports': ['usb'],
                      'appId': 'http://testserver'}
    u2f_request = {'appId': 'http://testserver',
                   'challenge': 'Listers_underpants',
                   'registeredKeys': [registered_key]}
    u2f_response = {
        'signatureData': ('AQAAABQwRQIhAOsMkVTyzVdHUeOvNJpCzUZjsHIs8vCJlmrwEwH90tR4AiAJHH-KB7OjUAoEemUkyiiyYd4QMK4YmTTr'
                          'dqCzn9U8YQ'),
        'clientData': ('eyAiY2hhbGxlbmdlIjogIkxpc3RlcnNfdW5kZXJwYW50cyIsICJvcmlnaW4iOiAiaHR0cDpcL1wvdGVzdHNlcnZlciIsICJ'
                       '0eXAiOiAibmF2aWdhdG9yLmlkLmdldEFzc2VydGlvbiIgfQ'),
        'keyHandle': key_handle}

    def test_no_user(self):
        with self.settings(LOGIN_URL='/login/'):
            response = self.client.get(self.url)

        self.assertRedirects(response, '/login/', fetch_redirect_response=False)

    def test_get(self):
        user = User.objects.create_user('kryten')
        session = self.client.session
        session[AUTHENTICATION_USER_SESSION_KEY] = user.pk
        session.save()

        response = self.client.get(self.url)

        self.assertContains(response, 'Authenticate Universal 2nd Factor (U2F) key')

    def test_post(self):
        user = User.objects.create_user('kryten')
        U2fDevice.objects.create(user=user, version='U2F_V2', key_handle=self.key_handle, public_key=self.public_key)
        session = self.client.session
        session[AUTHENTICATION_USER_SESSION_KEY] = user.pk
        session[AUTHENTICATION_REQUEST_SESSION_KEY] = self.u2f_request
        session.save()

        with self.settings(LOGIN_REDIRECT_URL='/redirect/'):
            response = self.client.post(self.url, {'u2f_response': json.dumps(self.u2f_response)})

        self.assertRedirects(response, '/redirect/', fetch_redirect_response=False)
        self.assertEqual(get_user(self.client), user)
        self.assertQuerysetEqual(U2fDevice.objects.values_list('user', 'counter'), [(user.pk, 20)], transform=tuple)
        self.assertNotIn(AUTHENTICATION_REQUEST_SESSION_KEY, self.client.session)
        self.assertNotIn(AUTHENTICATION_USER_SESSION_KEY, self.client.session)

    def test_post_no_session(self):
        # Test U2F request is not stored in session
        user = User.objects.create_user('kryten')
        session = self.client.session
        session[AUTHENTICATION_USER_SESSION_KEY] = user.pk
        session.save()

        response = self.client.post(self.url, {'u2f_response': 'null'})

        self.assertContains(response, 'Authenticate Universal 2nd Factor (U2F) key')
        self.assertEqual(response.context['form'].errors, {NON_FIELD_ERRORS: ['Authentication request not found.']})
        self.assertNotIn(AUTHENTICATION_REQUEST_SESSION_KEY, self.client.session)
        self.assertEqual(self.client.session[AUTHENTICATION_USER_SESSION_KEY], user.pk)

    def test_post_invalid_response(self):
        user = User.objects.create_user('kryten')
        self.client.force_login(user)
        session = self.client.session
        session[AUTHENTICATION_USER_SESSION_KEY] = user.pk
        session[AUTHENTICATION_REQUEST_SESSION_KEY] = self.u2f_request
        session.save()

        response = self.client.post(self.url, {'u2f_response': 'null'})

        self.assertContains(response, 'Authenticate Universal 2nd Factor (U2F) key')
        self.assertEqual(response.context['form'].errors, {NON_FIELD_ERRORS: ['Authentication failed.']})
        self.assertNotIn(AUTHENTICATION_REQUEST_SESSION_KEY, self.client.session)
        self.assertEqual(self.client.session[AUTHENTICATION_USER_SESSION_KEY], user.pk)



"""Test `django_fido.backends` module."""
from __future__ import unicode_literals

from django.contrib.auth import get_user_model
from django.contrib.messages.storage.cookie import CookieStorage
from django.core.exceptions import PermissionDenied
from django.test import RequestFactory, TestCase
from mock import sentinel

from django_fido.backends import U2fAuthenticationBackend
from django_fido.models import U2fDevice

User = get_user_model()


class TestU2fAuthenticationBackend(TestCase):
    """Test `U2fAuthenticationBackend` class."""

    backend = U2fAuthenticationBackend()

    # Valid U2F data
    key_handle = 'iYIO_N4276HND_X5IV8SP7Fi9bQcxdIeOHu2r9wf7RuFup9ywB-XwU18YkY0CWsH870-qbZdw7q5JuzyE4GqQQ'
    public_key = 'BHhw5KbmqBfVUNGiAZeyWxVbrBtUjThxwTeDcPmcG8hO6XHxp3bonp_OEVHkD601hTMIH6cReYLV1qBIyJ0NeIU'
    registered_key = {'publicKey': public_key, 'keyHandle': key_handle, 'version': 'U2F_V2', 'transports': ['usb'],
                      'appId': 'http://testserver'}
    u2f_request = {'appId': 'http://testserver',
                   'challenge': 'Listers_underpants',
                   'registeredKeys': [registered_key]}
    u2f_response = {
        'signatureData': ('AQAAABQwRQIhAOsMkVTyzVdHUeOvNJpCzUZjsHIs8vCJlmrwEwH90tR4AiAJHH-KB7OjUAoEemUkyiiyYd4QMK4YmTTr'
                          'dqCzn9U8YQ'),
        'clientData': ('eyAiY2hhbGxlbmdlIjogIkxpc3RlcnNfdW5kZXJwYW50cyIsICJvcmlnaW4iOiAiaHR0cDpcL1wvdGVzdHNlcnZlciIsICJ'
                       '0eXAiOiAibmF2aWdhdG9yLmlkLmdldEFzc2VydGlvbiIgfQ'),
        'keyHandle': key_handle}

    def test_authenticate(self):
        user = User.objects.create_user('kryten')
        U2fDevice.objects.create(user=user, version='U2F_V2', key_handle=self.key_handle, public_key=self.public_key)

        authenticated_user = self.backend.authenticate(sentinel.request, user, self.u2f_request, self.u2f_response)

        self.assertEqual(authenticated_user, user)
        self.assertQuerysetEqual(U2fDevice.objects.values_list('user', 'counter'), [(user.pk, 20)], transform=tuple)

    def test_authenticate_wrong_counter(self):
        user = User.objects.create_user('kryten')
        U2fDevice.objects.create(user=user, version='U2F_V2', key_handle=self.key_handle, public_key=self.public_key,
                                 counter=42)
        request = RequestFactory().get('/dummy/')
        request._messages = CookieStorage(request)

        self.assertRaisesMessage(PermissionDenied, "Counter didn't increase.",
                                 self.backend.authenticate, request, user, self.u2f_request, self.u2f_response)

        self.assertQuerysetEqual(U2fDevice.objects.values_list('user', 'counter'), [(user.pk, 42)], transform=tuple)

    def test_authenticate_invalid_request(self):
        self.assertIsNone(self.backend.authenticate(sentinel.request, sentinel.user, {}, self.u2f_response))

    def test_authenticate_invalid_response(self):
        self.assertIsNone(self.backend.authenticate(sentinel.request, sentinel.user, self.u2f_request, {}))

    def test_mark_device_used(self):
        user = User.objects.create_user('kryten')
        u2f_device = U2fDevice.objects.create(user=user, version='U2F_V2', key_handle='Left nipple', public_key='Turn')

        self.backend.mark_device_used(u2f_device, 42)

        self.assertQuerysetEqual(U2fDevice.objects.values_list('user', 'counter'), [(user.pk, 42)], transform=tuple)

    def test_mark_device_used_equal(self):
        # Test device returned the same counter.
        user = User.objects.create_user('kryten')
        u2f_device = U2fDevice.objects.create(user=user, version='U2F_V2', key_handle='Left nipple', public_key='Turn',
                                              counter=42)

        self.assertRaisesMessage(ValueError, "Counter didn't increase.", self.backend.mark_device_used, u2f_device, 42)

        self.assertQuerysetEqual(U2fDevice.objects.values_list('user', 'counter'), [(user.pk, 42)], transform=tuple)

    def test_mark_device_used_decrease(self):
        # Test device returned lower counter.
        user = User.objects.create_user('kryten')
        u2f_device = U2fDevice.objects.create(user=user, version='U2F_V2', key_handle='Left nipple', public_key='Turn',
                                              counter=42)

        self.assertRaisesMessage(ValueError, "Counter didn't increase.", self.backend.mark_device_used, u2f_device, 41)

        self.assertQuerysetEqual(U2fDevice.objects.values_list('user', 'counter'), [(user.pk, 42)], transform=tuple)

    def test_get_user(self):
        user = User.objects.create_user('kryten')
        self.assertEqual(self.backend.get_user(user.pk), user)

    def test_get_user_unknown(self):
        self.assertIsNone(self.backend.get_user(42))



"""Authentication backends for Django FIDO."""
from __future__ import unicode_literals

import logging

from django.contrib import messages
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.utils.translation import ugettext_lazy as _
from u2flib_server import u2f

_LOGGER = logging.getLogger(__name__)


class U2fAuthenticationBackend(object):
    """
    Authenticate user using U2F.

    @cvar counter_error_message: Error message in case U2F device counter didn't increase.
    """

    counter_error_message = _("Counter of the U2F device decreased. Device may have been duplicated.")

    def authenticate(self, request, user, u2f_request, u2f_response):
        """Authenticate using U2F."""
        try:
            device, counter, user_presence = u2f.complete_authentication(u2f_request, u2f_response)
        except (TypeError, ValueError, KeyError) as error:
            _LOGGER.info("U2F authentication failed with error: %r", error)
            return None

        u2f_device = user.u2f_devices.get(key_handle=device['keyHandle'])
        try:
            self.mark_device_used(u2f_device, counter)
        except ValueError:
            # Raise `PermissionDenied` to stop the authentication process and skip remaining backends.
            messages.error(request, self.counter_error_message)
            raise PermissionDenied("Counter didn't increase.")
        return user

    def mark_device_used(self, u2f_device, counter):
        """Update U2F device usage information."""
        if counter <= u2f_device.counter:
            _LOGGER.info("U2F authentication failed because of not increasing counter.")
            raise ValueError("Counter didn't increase.")
        u2f_device.counter = counter
        u2f_device.full_clean()
        u2f_device.save()

    def get_user(self, user_id):
        """Return user based on its ID."""
        try:
            return get_user_model().objects.get(pk=user_id)
        except get_user_model().DoesNotExist:
            return None



"""Test `django_fido.forms` module."""
from __future__ import unicode_literals

from django.test import SimpleTestCase

from django_fido.forms import U2fResponseForm


class TestU2fResponseForm(SimpleTestCase):
    """Test `U2fResponseForm` class."""

    def test_clean_u2f_response(self):
        form = U2fResponseForm({'u2f_response': '{"answer": 42}'})

        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data, {'u2f_response': {'answer': 42}})

    def test_clean_u2f_response_invalid(self):
        form = U2fResponseForm({'u2f_response': 'JUNK'})

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {'u2f_response': ['U2F response is malformed.']})
